<html>
<head>
  <title>Channel: {{ channel }}</title>
  <link rel="stylesheet" href="style.css"/>
</head>
<body>
  <div class="header">
    <a href="./index.html"><img src="icon.png" width="32px" height="32px" alt="home"/></a>
    &nbsp;
    Anaconda Mirror
  </div>
  <h2>
     Channel: <a href="./index.html">{{ channel }}</a><br/>
  </h2>
  <table>
    <tr class="channeldata">
      <th>Package</th><th>Latest<br/>Version</th>
      {% for subdir in subdirs %}<th class="subdir"><a href="{{ subdir }}/index.html">{{ subdir | hyphenate }}</a></th>{% endfor %}
      <th>Last Modified</th>
    </tr>
{% for record in packages %}
    <tr>
      <td>{{ record.name }}</td><td>{{ record.version }}</td>
      {%- for subdir in subdirs %}<td>{% if subdir in record.subdirs %}X{% endif %}</td>{% endfor -%}
      <td>{% if record.timestamp %}{{ record.timestamp|strftime("%Y-%m-%d %H:%M:%S") }}{% endif %}</td>
    </tr>
{% endfor %}
  </table>
  <address>
    Created by anaconda-mirror {{ mirror_version }}
    <br />
    Updated: {{ current_time|strftime("%Y-%m-%d %H:%M:%S %z") }}
    <br/>
    Records: {{ packages|length }}
  </address>
</body>
</html>
