html {
  font-size: 62.5%;
}
body {
  background-color: #f2f2f2;
  font-size: 1.6rem;
}
/* Using pixels here to provide
   good alignment with the logo */
div.header {
  display: flex;
  align-items: center;
  font-size: 24px;
  font-weight: bold;
  margin-bottom: 16px;
}
h2 {
  font-size: 1.25em;
  font-weight: bold;
  margin-bottom: 1em;
}
a, a:active, a:visited {
  text-decoration: none;
  color: blue;
}
a:hover, a:focus {
  text-decoration: underline;
  color: red;
}
pre {
    width: min-content;
    background: #dddddd;
    padding: 1em;
    margin-left: 2em;
    border: 1px solid #aaaaaa;
}
table {
  border-collapse: collapse;
  background-color: white;
  border-top: 1px solid #888888;
  border-bottom: 1px solid #888888;
  margin-top: 1em;
  font-size: 1em;
  white-space: nowrap;
}
table.small {
  font-size: 0.75em;
}
tr.firstpkg > td {
  border-top: 1px solid #e0e0e0;
}
th, td {
  font: 100% monospace;
  text-align: left;
}
td {
  padding: 0.2em 0.5em;
}
tr.channeldata th {
  font-size: 80%;
}
th {
  font-weight: bold;
  text-align: center;
  padding: 0.5em 0.5em;
}
tr:last-child > td {
  padding-bottom: 0.5em;
}
.fqdn {
  font: 100% monospace;
  font-weight: bold;
}
address {
  color: #787878;
  padding-top: 1em;
  font-size: 0.8em;
}
