<html>
<head>
  <title>Subdir: {{ channel }} / {{ subdir }}</title>
  <link rel="stylesheet" href="../style.css"/>
</head>
<body>
  <div class="header">
    <a href="../index.html"><img src="../icon.png" width="32px" height="32px" alt="logo"/></a>
    &nbsp;
    Anaconda Mirror
  </div>
  <h2>
    Channel: <a href="../index.html">{{ channel }}</a><br />
    Platform: <a href="./index.html">{{ subdir }}</a>
  </h2>
  <table class="small">
    <tr>
      <th>Filename</th>
      <th>Size</th>
      <th>Last Modified</th>
      <th>MD5</th>
      <th>SHA256</th>
    </tr>
{% for path in extra_paths %}
    {% set record = extra_paths[path] %}
    <tr>
      <td>{{ path | add_href(path) }}</td>
      <td class="s">{{ record.size | human_bytes }}</td>
      <td>{% if record.timestamp %}{{ record.timestamp|strftime("%Y-%m-%d %H:%M:%S") }}{% endif %}</td>
      <td>{{ record.md5 }}</td>
      <td>{{ record.sha256 }}</td>
    </tr>
{%- endfor %}
{% for fn, record in packages %}
    <tr>
      <td>{{ fn | add_href(fn) }}</td>
      <td class="s">{{ record.size | human_bytes }}</td>
      <td>{% if record.timestamp %}{{ record.timestamp|strftime("%Y-%m-%d %H:%M:%S") }}{% endif %}</td>
      <td>{{ record.md5 }}</td>
      <td>{{ record.sha256 }}</td>
    </tr>
{%- endfor %}
  </table>
  <address>
    Created by anaconda-mirror {{ mirror_version }}
    <br />
    Last updated: {{ current_time|strftime("%Y-%m-%d %H:%M:%S") }}
    <br/>
    Packages: {{ packages|length }}
  </address>
</body>
</html>
