



anaconda-mirror-repo4 --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
anaconda-mirror-ae5 --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
anaconda-mirror-as6 --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
anaconda-mirror-local --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
anaconda-mirror-s3 --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
